/*
=head1 TreeNode.cc - member routines for class TreeNode

	Copyright 1994 Allan Brighton.

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies.  
Allan Brighton make no representations about the suitability of this 
software for any purpose. It is provided "as is" without express or 
implied warranty.


The basic algorithm used below is taken from:

    IEEE Software: July 1990 "Drawing Dynamic Trees"

I also used the Xview extension "Slingshot" as a reference.
*/
 

#include <stdlib.h>
#include <string.h>
#include "TreeNode.h"

inline int max(int x, int y) { return (x > y) ? x : y; }

    
// =head1 TreeNode - constructor
// =head2 Synopsis
// 	TreeNode::TreeNode()
// =head2 Description
// This is the constructor for a new child node. It sets up the
// data structure and initializes it with the values given.

// =head2 Arguments
// =over 3
// =item const char *tag
// name of this node
// =item int x, int y
// position of the node (will be computed if -1)
// =item int w, int h
// width and height of the node
// =back

// =head2 Return Value
// the constructor returns the pointer to the new node
*/
TreeNode::TreeNode(const char* tag, int x, int y, int w, int h, int border) 
: tag_(strdup(tag)), 
  parent_(NULL), 
  child_(NULL), 
  sibling_(NULL), 
  width_(w), 
  height_(h), 
  border_(border), 
  pos_(x, y)
{
}


// =head1 TreeNode::box - setup box coords
// =head2 Synopsis
//	TreeNode::box(int x1, int y1, int x2, int y2)
// =head2 Description
// set the bounding box of a node.
// =head2 Arguments
// =over 3
// =item x1, y1, x2, y2
// Cooridinates of the new bounding box of this node
// =back

void TreeNode::box(int x1, int y1, int x2, int y2)
{
    pos_.x = x1;
    pos_.y = y1;
    width_ = x2 - x1;
    height_ = y2 - y1;
}


/*
=head1 TreeNode destructor
The destructor cleans up the mess left over by a single node
*/
TreeNode::~TreeNode()
{
    free(tag_);
       
    if (child_ != NULL) {
	delete child_;
	child_ = NULL;
    }
        
    if (sibling_ != NULL) {
	delete sibling_;
	sibling_ = NULL;
    }
}


/*
=head1 newTreePolyline - inter-node connection
=head2 Synopsis
	TreeNode::newTreePolyline(int x, int y, TreePolyline *pl)

Keep track of the allocated polylines for later garbage collection
*/
TreePolyline* TreeNode::newTreePolyline(int x, int y, TreePolyline* pl)
{
    TreePolyline* p = new TreePolyline(x, y, pl);
    treePolylines_.append(p);
    return p;
}



/*
=head1 setXY - set position
=head2 Synopsis
	TreeNode::setXY(int &x1, int &y1, int &x2, int &y2)

=head2 Description
Set the absolute positions of the tree nodes from the relative
offsets set by C<layout()>.

The C<offset_> field is relative to the nodes predecessor 
(parent or sibling). From that we want to set the C<pos_> field.

Assume that this node already has the correct pos.

C<x1, y1, x2, y2> are set to the bounding box of the entire tree

*/
void TreeNode::setXY(int& x1, int& y1, int& x2, int& y2)
{
    if (child_ != NULL) 
	child_->setPosFrom(this, x1, y1, x2, y2);
        
    if (sibling_ != NULL) 
	sibling_->setPosFrom(this, x1, y1, x2, y2);
}


/*
 * Set this node's position from its relative offset and its predecessor's
 * position.
 *
 * x1, y1, x2, y2 are kept up to date as the bounding box of the tree
*/
void TreeNode::setPosFrom(TreeNode* pred, int& x1, int& y1, int& x2, int& y2)
{
    prev_pos_ = pos_;
    if (horizontal())
        pos_ = pred->pos_ + offset_;
    else {
        pos_.x = pred->pos_.x + offset_.y;
        pos_.y = pred->pos_.y + offset_.x;
    }
    
    // keep track of the outer bounds of the tree
    if (pos_.x < x1)
        x1 = pos_.x;
    if (pos_.y < y1)
        y1 = pos_.y;
    if ((pos_.x+width_) > x2)
        x2 = pos_.x+width_;
    if ((pos_.y+height_) > y2)
        y2 = pos_.y+height_;
        
    setXY(x1, y1, x2, y2);
}


/*
 * insert the node in the tree as a child of this node
*/
void TreeNode::addLink(TreeNode* node)
{
    node->parent_ = this;
    if (child_ == NULL)
        child_ = node;
    else {
       /* risc: gcc-2.7.0 has other scoping */
       TreeNode* p = child_;
       for (; p->sibling_; p = p->sibling_);
       p->sibling_ = node;
    }
}


/*
=head1 merge - merge nodes
=head2 Synopsis
	int TreeNode::merge(TreePolygon &p1, TreePolygon &p2)

=head2 Description
C<merge()> walks through two contours and prevents overlap. The lower
contour of the upper node is compared against the upper contour of the
lower node.

See the article mentioned at the top of this file for details on
what this function does.

=head2 Return Value
this function returns the total offset needed by the 
subpolygon.
*/
int TreeNode::merge(TreePolygon& p1, TreePolygon& p2)
{
    int     x, y, total, d;
    TreePolyline *lower, *upper, *b;

    x = y = total = 0;
    upper = p1.lower.head;
    lower = p2.upper.head;

    while (lower && upper) {	/* compute offset total */

	d = offset(x, y, lower->p.x, lower->p.y, upper->p.x, upper->p.y);
	y += d;
	total += d;

	if (x + lower->p.x <= upper->p.x) {
	    y += lower->p.y;
	    x += lower->p.x;
	    lower = lower->link;
	}
	else {
	    y -= upper->p.y;
	    x -= upper->p.x;
	    upper = upper->link;
	}
    }
    
    /* store result in c1 */
    if (lower) {
	b = bridge(p1.upper.tail, 0, 0, lower, x, y);
	p1.upper.tail = (b->link) ? p2.upper.tail : b;
	p1.lower.tail = p2.lower.tail;
    }
    else {
	b = bridge(p2.lower.tail, x, y, upper, 0, 0);
	if (b->link == NULL)
	    p1.lower.tail = b;
    }
    p1.lower.head = p2.lower.head;

    return total;
}











